_         = require("lodash")
Bluebird  = require("bluebird")
cert      = require("@packages/https-proxy/test/helpers/certs")
https     = require("https")
useragent = require("express-useragent")
{ allowDestroy } = require("@packages/network")
e2e       = require("../support/helpers/e2e")

## create an HTTPS server that forces TLSv1
startTlsV1Server = (port) ->
  Bluebird.fromCallback (cb) ->
    opts = _.merge({
      secureProtocol: "TLSv1_server_method",
    }, cert)

    serv = https.createServer opts, (req, res) =>
      res.setHeader('content-type', 'text/html')
      res.end('foo')

    allowDestroy(serv)

    serv.listen port, (err) =>
      cb(null, serv)

    serv.on('error', cb)

onServer = (app) ->
  app.get "/agent.json", (req, res) ->
    source = req.headers["user-agent"] ? ""

    ua = useragent.parse(source)

    res.send({agent: ua})

  app.get "/agent.html", (req, res) ->
    source = req.headers["user-agent"] ? ""

    ua = useragent.parse(source)

    res.send("""
      <html>
        <a href='/agent.html'>agent</a>
        <div id="agent">
          #{JSON.stringify(ua)}
        </div>
      </html
    """)

  app.get "/headers.html", (req, res) ->
    res.send("""
      <html>
        <div id="headers">
          #{JSON.stringify(req.headers)}
        </div>
      </html>
    """)

  app.get "/fail", (req, res) ->
    res.sendStatus(500)

  app.get "/timeout", (req, res) ->
    ms = req.query.ms ? 0

    setTimeout ->
      res.send("<html>timeout: <span>#{ms}</span></html>")
    , ms

  app.get "/response_never_finishes", (req, res) ->
    ## dont ever end this response
    res.type("html").write("foo\n")

  ## https://github.com/cypress-io/cypress/issues/5602
  app.get "/invalid-header-char", (req, res) ->
    ## express/node may interfere if we just use res.setHeader
    res.connection.write(
        """
        HTTP/1.1 200 OK
        Content-Type: text/html
        Set-Cookie: foo=bar-#{String.fromCharCode(1)}-baz

        foo
        """
    )

    res.connection.end()

describe "e2e visit", ->

  context "low response timeout", ->
    e2e.setup({
      settings: {
        responseTimeout: 500
        pageLoadTimeout: 1000
      }
      servers: {
        port: 3434
        static: true
        onServer: onServer
      }
    })

    e2e.it "passes", {
      spec: "visit_spec.coffee"
      snapshot: true
      onRun: (exec) ->
        startTlsV1Server(6776)
        .then (serv) ->
          exec()
          .then ->
            serv.destroy()
    }

    e2e.it "passes with experimentalSourceRewriting", {
      spec: "source_rewriting_spec.js"
      config: {
        experimentalSourceRewriting: true
      }
      snapshot: true
      onRun: (exec) ->
        startTlsV1Server(6776)
        .then (serv) ->
          exec()
          .then ->
            serv.destroy()
    }

    e2e.it "fails when network connection immediately fails", {
      spec: "visit_http_network_error_failing_spec.coffee"
      snapshot: true
      expectedExitCode: 1
    }

    e2e.it "fails when server responds with 500", {
      spec: "visit_http_500_response_failing_spec.coffee"
      snapshot: true
      expectedExitCode: 1
    }

    e2e.it "fails when file server responds with 404", {
      spec: "visit_file_404_response_failing_spec.coffee"
      snapshot: true
      expectedExitCode: 1
    }

    e2e.it "fails when content type isnt html", {
      spec: "visit_non_html_content_type_failing_spec.coffee"
      snapshot: true
      expectedExitCode: 1
    }

    e2e.it "calls onBeforeLoad when overwriting cy.visit", {
      snapshot: true
      spec: "issue_2196_spec.coffee"
    }

  context "low responseTimeout, normal pageLoadTimeout", ->
    e2e.setup({
      settings: {
        responseTimeout: 2000
      }
      servers: {
        port: 3434,
        static: true,
        onServer: onServer
      }
    })

    e2e.it "fails when response never ends", {
      spec: "visit_response_never_ends_failing_spec.js",
      snapshot: true,
      expectedExitCode: 3
    }

  context "normal response timeouts", ->
    e2e.setup({
      settings: {
        pageLoadTimeout: 1000
      }
      servers: {
        port: 3434
        static: true
        onServer: onServer
      }
    })

    e2e.it "fails when visit times out", {
      spec: "visit_http_timeout_failing_spec.coffee"
      snapshot: true
      expectedExitCode: 2
    }
